/*
 * libtg-example.cpp - Small example demonstrating the use of the libtg API
 *
 *  Created on: Jan 28, 2012
 *      Author: botond
 */

#include <vector>
#include <string>
#include <iostream>
#include <functional>
#include <algorithm>
#include <cassert>

#include <tg/api.hpp>

using std::vector;
using std::string;
using std::function;
using std::find_if;
using std::cout;

using tg::TimetableManager;
using tg::session_id_t;
using tg::StudySession;
using tg::CourseList;
using tg::session_type;
using tg::GenerationStatistics;
using tg::ProgressIndicator;

int main(int argc, const char** argv)
{
    // Create the TimetableManager singleton object that is used to access
    // API functionality.
    TimetableManager& mgr = TimetableManager::Create("sessions.txt", "", argv[0]);
    
    // List all the sessions of study that the library knows about 
    vector<StudySession> study_sessions = mgr.GetStudySessions();
    cout << "Here are the available study sessions:\n";
    for (const auto& study_session : study_sessions)
    {
        cout << "\tName: " << study_session.name << "\n";
        cout << "\tFaculties: " << study_session.faculty_list << "\n\n";
    }
    
    // Start a user session
    session_id_t session_id = mgr.StartSession();
    
    // Select a study session
    size_t selected = find_if(study_sessions.begin(), study_sessions.end(),
                              [](const StudySession& s)
                              {
                                  return s.name.find("Fall/Winter") != string::npos
                                      && s.faculty_list.find("Arts and Science") != string::npos;
                              })
                      - study_sessions.begin();
    assert(selected < study_sessions.size());
    mgr.SelectStudySession(session_id, selected);
    
    // List the courses available in the selected study session
    CourseList courses = mgr.GetCoursesForSelectedStudySession(session_id);
    cout << "Here are the courses available in this study session:\n";
    for (const auto& department : courses.departments)
        for (const auto& course : department.second)
        {
            cout << course.get_activity_code_and_title() << "\n";
            cout << "Offered in sessions:\n";
            for (session_type session : course.sessions_offered)
                cout << "\t" << ( session == session_type::F ? "fall"
                                : session == session_type::S ? "spring"
                                :                              "full-year") << "\n";
            cout << "\n";
        }
    
    // Select some courses
    mgr.SelectCourses(session_id, {"CSC108H1", "CSC148H1"});
    
    // Specify the semesters in which we want to take the courses
    mgr.SpecifyTerm(session_id, "CSC108H1", session_type::F);
    mgr.SpecifyTerm(session_id, "CSC148H1", session_type::S);
    
    // Generation possible configurations (a "configuration" is a possible
    // placement of half-courses into fall or spring semester). 
    // Since we specified semeesters for all the courses, there will be only
    // one, so it is auto-selected. Otherwise we'd have to call 
    // SelectConfigurations() before calling GenerateTimetables().
    mgr.GenerateConfigurations(session_id);
    
    // Generate timetables for the selected configurations.
    // This takes callbacks for indicating progress and reporting generation
    // statistics, but we pass no-ops for those.
    mgr.GenerateTimetables(session_id, 
                           ProgressIndicator{[](int){}, 100}, // no-op progress indicator 
                           [](GenerationStatistics){});  // no-op gen stats callback
    
    // See how many timetables were generated
    cout << "Number of possible timetables: " << mgr.GetTimetableCount(session_id, 0) << "\n";
    
    // End the user session
    mgr.EndSession(session_id);
}
