// Timetable Generator API Include File
// 
// Copyright Botond Ballo 2012. Use, modification and distribution
// is subject to the Timetable Generator API License, version 1.0
// (see the accompanying file LICENSE.txt).

#ifndef LIBTG_INCLUDE_TG_API_HPP_
#define LIBTG_INCLUDE_TG_API_HPP_

#include <memory>      // std::unique_ptr
#include <string>      // std::string
#include <ostream>     // std::ostream
#include <vector>      // std::vector
#include <utility>     // std::move, std::pair
#include <map>         // std::map
#include <functional>  // std::function
#include <istream>     // std::istream
#include <ostream>     // std::ostream

#include <cstdint>  // uint64_t

namespace tg
{
    namespace detail
    {
    
        using std::unique_ptr;
        using std::string;
        using std::ostream;
        using std::vector;
        using std::move;
        using std::pair;
        using std::map;
        using std::function;
        using std::istream;
        using std::ostream;
        
        // Public types
        // TODO: MSVC 10 compatibility (doesn't support enum classes or forward-declaring enums)
        
        typedef uint64_t session_id_t;
        
        enum class term_type {fall, spring};
        
        enum class session_type {F = 0, S, Y};
        
        enum class timetable_sort_criterion : size_t;
        
        enum class course_load_state {empty, session_selected, courses_selected, configs_selected, timetables_generated};
        
        typedef vector<timetable_sort_criterion> timetable_sort_criteria;
        
        // a "class_time" is a time when a meeting starts or ends
        struct class_time
        {
            // constructors
            class_time() : hour(), half() {}  // ugh, default constructor 
            class_time(unsigned hour, bool half) : hour(hour), half(half) {}

            // overloaded operators
            void operator++();  // increment by 30 minutes
            void operator--();  // decrement by 30 minutes

            // text output
            string to_str() const;

            // data members
            unsigned hour;  // can range from 9 to 21
            bool     half;  // add 30 minutes to "hour"
        };
        
        // utilities for class_time
        
        // find the number of unit times between a and b
        size_t operator-(const class_time& a, const class_time& b);
        
        // Represents one time block (part of the GUI-suitable timetable representation)
        struct block
        {
            // constructor
            // TODO: this constructor will be obsoleted by C++0x uniform initialization
            block(unsigned day, const class_time& start_time, size_t length)
                : day(day), start_time(start_time), length(length) {}

            // data members
            unsigned   day;         // day of week (0 = Monday, 4 = Friday)
            class_time start_time;  // start time of block
            size_t     length;      // length of block in unit blocks (currently 1 unit block = 30 min)
        };

        // Time blocks for one course in a page of a timetable (part of the GUI-suitable timetable representation)
        struct course_blocks
        {
            // constructor
            // TODO: this constructor will be obsoleted by C++0x uniform initialization
            course_blocks(const string& activity_code, const string& section_code, const string& instructor,
                          const string& instructor_short, vector<block>&& blocks)
                        : activity_code(activity_code), section_code(section_code), instructor(instructor),
                          instructor_short(instructor_short), blocks(move(blocks)) {}

            // data members
            string        activity_code;     // activity code for these blocks
            string        section_code;      // section code for these blocks
            string        instructor;        // instructor for these blocks
            string        instructor_short;  // shortened version of 'instructor' 
            vector<block> blocks;            // the time blocks themselves
        };

        // This is one page in the GUI-suitable page representation of a timetable
        struct timetable_page
        {
            // constructor
            // TODO: this constructor will be obsoleted by C++0x uniform initialization
            timetable_page(term_type term, bool alternate, unsigned week, vector<course_blocks>&& courses)
                         : term(term), alternate(alternate), week(week), courses(move(courses)) {}
            
            // convenience accessors
            string name() const;

            // data members
            term_type             term;       // fall or spring term
            bool                  alternate;  // is this page specific to one of the alternating weeks?
            unsigned              week;       // if so, which week?
            vector<course_blocks> courses;    // the blocks of time in this timetable, categorized by course
                                              // i.e. one element of this vector stores the blocks for one course
        };

        // This is a representation of a timetable that is suitable for the GUI display to use
        struct timetable_page_rep
        {
            // constructor
            // TODO: this constructor will be obsoleted by C++0x uniform initialization
            //       (if we make the data members public)
            timetable_page_rep(vector<timetable_page>&& pages) : pages(move(pages)) {}

            // accessors
            const timetable_page& operator[](size_t i) const { return pages[i]; }
            size_t num_pages() const { return pages.size(); }
            
            // static methods
            static size_t num_pages(bool has_alternate_weeks);
        private:
            // data members
            vector<timetable_page> pages;
        };
        
        // Represents a single offering a course
        struct Offering
        {
            // constructor
            // TODO: this constructor will be obsoleted by C++0x uniform initialization
            Offering(const string& activity_code, session_type session) 
                   : activity_code(activity_code), session_code(session) {}
            
            string       activity_code;
            session_type session_code;
            
            string activity_session_code() const;
            
            bool operator==(const Offering& other) const
            {
                return activity_code == other.activity_code && session_code == other.session_code;
            }
        };
        
        // Represents all the offerings of a course in a session of study
        struct Offerings
        {
            // constructor
            // TODO: this constructor will be obsoleted by C++0x uniform initialization
            Offerings(const string& activity_code, const string& title, vector<session_type>&& sessions_offered)
                    : activity_code(activity_code), title(title), sessions_offered(sessions_offered) {}
            
            string               activity_code;
            string               title;
            vector<session_type> sessions_offered;
            
            string get_activity_code_and_title() const;
            
            // Get the activity code portion of the combined (activity code + title)
            // returned by get_activity_code_and_title().
            static string get_activity_code_part(const string& activity_code_and_title);
            
            // Get the department code of a course from its activity code.
            // This can be passed either the activity code itself, or the string
            // returned by get_activity_code_and_title().
            static string get_department_code(const string& activity_code);
        };
        
        // Represents all the offerings of all courses in a session of study
        struct CourseList
        {
            // constructor
            // TODO: this constructor will be obsoleted by C++0x uniform initialization
            CourseList(map<string, vector<Offerings>>&& departments) : departments(departments) {}
            
            map<string, vector<Offerings>> departments;  // department name -> courses in department
        };
        
        // Represents information about a single activity configuration
        struct ActivityConfiguration
        {
            // constructor
            // TODO: this constructor will be obsoleted by C++0x uniform initialization
            ActivityConfiguration(vector<Offering>&& offerings, bool has_alternate_timeslots)
                                : offerings(move(offerings)), 
                                  has_alternate_timeslots(has_alternate_timeslots) {}
            
            vector<Offering> offerings;                // offerings in configuration
            bool             has_alternate_timeslots;  // does this configuration need separate pages for alternate weeks?
            
            bool operator==(const ActivityConfiguration& other) const
            {
                return has_alternate_timeslots == other.has_alternate_timeslots && offerings == other.offerings;
            }
        };
        
        // Represents information about a session of study
        struct StudySession
        {
            // constructor
            // TODO: this constructor will be obsoleted by C++0x uniform initialization
            StudySession(const string& name, const string& faculty_list, bool is_default) : name(name), faculty_list(faculty_list), is_default(is_default) {}
            
            string name;          // name of session (e.g. "2010 Summer")
            string faculty_list;  // faculties available in this study session
            bool   is_default;    // is this the default study session?
        };
        
        // A class that stores various statistics about the timetable generation process
        // which may be displayed to the user during generation.
        struct GenerationStatistics
        {
            // constructor
            GenerationStatistics() : timetables_generated(0), average_generation_speed(0) {}
            
            // data members
            size_t   timetables_generated;      // number of timetables generated so far
            string   elapsed_time;              // elapsed time since beginning of generation processed
            unsigned average_generation_speed;  // in timetables per second
        };
        
        // Stores the name of the file where a course load is saved
        // (if any), and whether the course load has changed since
        // the last time it was saved to this file.
        struct FileSaveStatus
        {
            // constructor
            // TODO: this constructor will be obsoleted by C++0x uniform initialization
            FileSaveStatus(const string& filename, bool up_to_date) : filename(filename), up_to_date(up_to_date) {}
            
            string filename;
            bool   up_to_date;
        };
        
        // TimetableDataSource and TimetableDataSink are abstractions of a "data store"
        // from which timetable data can be read and to which it can be written, respectively.
        // The interface for reading is to get an istream corresponding to a given "key".
        // The interface for writing is to get an ostream corresponding to a given key.
        // The most obvious implementation is a filesystem, with the streams being file streams
        // and the keys being filenames.
        struct TimetableDataSource
        {
            virtual ~TimetableDataSource() {}
            virtual istream& get_reader(const string& key) = 0; 
        };
        struct TimetableDataSink
        {
            virtual ~TimetableDataSink() {}
            virtual ostream& get_writer(const string& key) = 0;
        };
        
        // Used to report the progress of time consuming API calls.
        // 'callback' is a function the API call will call from time to time,
        // with an integer argument 'arg' in the range [0, 'granularity'], to
        // indicate that 'arg'/'granularity' of the process is complete.
        // The callback will be called no more than 'granularity' + 1 times
        // in total.
        struct ProgressIndicator
        {
            function<void(int)> callback;
            size_t              granularity;
        };

        // Utility functions
        // These are part of the API
        // They are completely stateless, requiring neither per-session nor global state
        
        // Get the earliest start time for a class
        class_time GetEarliestClassTime();
        
        // Get the total number of course blocks from morning to evening
        size_t GetTotalTimeBlocks();
        
        // Get a list of possible criteria to be used when sorting timetables, along with their descriptions.
        // Note: the list of criteria depends on whether or not the activity configuration whose timetables
        // are being sorted has alternate timeslots. 
        map<timetable_sort_criterion, string> GetTimetableSortCriteria(bool has_alternate_timeslots);
        
        // TimetableManager class
        // This is the gateway to the API functions.
        
        class timetable_manager_impl;
        
        class TimetableManager
        {
        public:
            // Public interface.
            // This *is* the API.
            
            // Make class non-copyable
            TimetableManager(const TimetableManager&) = delete;
            
            // Creates the (only) timetable_manager instance.
            static TimetableManager& Create(const string& study_sessions_file, 
                                            const string& course_data_filename_prefix = "", 
                                            const string& argv0 = "");

            // Get a list of all study sessions
            vector<StudySession> GetStudySessions() const;
            
            // Get the courses available for a given study session
            CourseList GetCoursesForStudySession(size_t study_session_index) const;

            // Check connectivity to the university websites from which timetable data is downloaded
            bool CheckUniversityWebsiteConnectivity() const;

            // Get a list of the filenames of the course data files
            vector<string> GetCourseDataFileNames() const;
            
            // Update timetable data from the university websites
            void UpdateTimetableData(ProgressIndicator progress_indicator);
            
            // Load timetable data from the specified source
            void LoadTimetableData(TimetableDataSource& source, ProgressIndicator progress_indicator);
            
            // Reload all data used by the API. This includes inventories as well as other data
            // like study sessions. This should be called when a data file like sessions.txt or
            // altwkexcept.txt changes
            void ReloadData();
            
            // Start a new session
            session_id_t StartSession();
            
            // End a session
            void EndSession(session_id_t session_id);

            // Check whether a course load is open
            bool IsCourseLoadOpen(session_id_t session_id) const;
            
            // Determine whether a course load has been modified since
            // the last time it was modified (false), or if it is up to date (true).
            bool IsCourseLoadUpToDate(session_id_t session_id) const;
            
            // Get the name of the file the active course load has last been saved to.
            // Returns the empty stirng if the active course load has never been saved.
            string GetCourseLoadFileName(session_id_t session_id) const;
            
            // Save the active course load to the given file.
            void SaveCourseLoad(session_id_t session_id, const string& filename);
            
            // Open a course load from a file.
            void OpenCourseLoad(session_id_t session_id, const string& filename);
            
            // Reopen the active course load.
            // This puts the course load into the same state as it would be in
            // had it been saved to a file and re-loaded from the file, but without
            // actually saving it to a file.
            void ReopenCourseLoad(session_id_t session_id);

            // Register a callback function to be called whenever the course load's
            // file save status changes.
            void ObserveFileSaveStatus(session_id_t session_id, function<void(FileSaveStatus)> callback);
            
            // Get the current course load state
            course_load_state GetCourseLoadState(session_id_t session_id) const;
            
            // Get the currently selected study session
            string GetCurrentStudySession(session_id_t session_id) const;

            // Select a study session
            void SelectStudySession(session_id_t session_id, size_t study_session_index);
            
            // Clear the selected study session
            void ClearStudySession(session_id_t session_id);
            
            // Get the courses available for the currently selected study session
            CourseList GetCoursesForSelectedStudySession(session_id_t session_id) const;
            
            // Get a list of currently selected courses (without terms being specified)
            vector<Offerings> GetSelectedCourses(session_id_t session_id) const;
            
            // Select courses (without specifying terms)
            void SelectCourses(session_id_t session_id, const vector<string>& activity_codes);
            
            // Clear selected courses
            void ClearSelectedCourses(session_id_t session_id);

            // For a selected course, specify what term it is taken in
            void SpecifyTerm(session_id_t session_id, const string& activity_code, session_type term);
            
            // Generate configurations, taking into account terms specified using SpecifyTerm.
            // Returns whether the selections made using SpecifyTerm are enough to decide
            // which configurations theuser wants to generate timetables for (true), or it is
            // necessary to ask the user to select among the configurations generated by
            // the configuration generation algorithm (false).
            // In the former case, the chosen configurations are auto-selected, so there is
            // no need to call SelectConfigurations().
            bool GenerateConfigurations(session_id_t session_id);
            
            // Un-specify the hints that were given using SpecifyTerm and applied using
            // ApplyConfigurationHints.
            void UnspecifyTerms(session_id_t session_id);
            
            // Get the number of configurations in the active course load.
            size_t GetConfigCount(session_id_t session_id) const;
            
            // Select configurations in the active course load
            void SelectConfigurations(session_id_t session_id, const vector<size_t>& configs);
            
            // Get the number of selected configurations in the active course load
            size_t GetSelectedConfigCount(session_id_t session_id) const;
            
            // Get the number of activities in the active course load
            size_t GetActivityCount(session_id_t session_id) const;
            
            // Get information about an available configuration in the active course load
            ActivityConfiguration GetConfig(session_id_t session_id, size_t config_index) const;
            
            // Get information about a selected configuration in the active course load
            ActivityConfiguration GetSelectedConfig(session_id_t session_id, size_t config_index) const;
            
            // Clear selected configurations
            void ClearSelectedConfigurations(session_id_t session_id);
            
            // Generate timetables for a given configuration
            void GenerateTimetables(session_id_t session_id, ProgressIndicator progress_indicator,
                                    function<void(GenerationStatistics)> generation_stats_callback);
            
            // Get the number of timetables for a given configuration
            size_t GetTimetableCount(session_id_t session_id, size_t config_index) const;
            
            // Get a timetable
            // The timetable lives inside the backend; it may be, doesn't have to be, copied.
            const timetable_page_rep& GetTimetable(session_id_t session_id, size_t config_index, size_t timetable_index) const;
            
            // Clear the generated timetables
            void ClearTimetables(session_id_t session_id);
            
            // Get pairs of conflicting meeting sections for a given configuration
            vector<pair<string, string>> GetConflicts(session_id_t session_id, size_t config_index) const;
            
            // Get a list of criteria by which timetables for the given configuration are currently sorted.
            timetable_sort_criteria GetCurrentSortCriteria(session_id_t session_id, size_t config_index) const;
            
            // Sort the timetables for a particular configuration, with the given sort criteria.
            // Use the given callback function as a progress indicator, which has the given display granularity.
            void SortTimetables(session_id_t session_id, size_t config_index, const timetable_sort_criteria& sort_criteria, 
                                ProgressIndicator progress_indicator);
            
            // Clear course load of all state
            void ClearCourseLoad(session_id_t session_id);
        private:
            // Constructor/destructor.
            // These are private because this class is a singleton.
            TimetableManager(timetable_manager_impl*);
            ~TimetableManager();
            
            // Static variables
            static bool m_created;
            
            // Instance variables
            unique_ptr<timetable_manager_impl> impl;
        };
    
    }  // end namespace detail
    
    using detail::session_id_t;
    using detail::session_type;
    using detail::timetable_sort_criterion;
    using detail::timetable_sort_criteria;
    using detail::course_load_state;
    using detail::class_time;
    using detail::block;
    using detail::course_blocks;
    using detail::timetable_page;
    using detail::timetable_page_rep;
    using detail::Offering;
    using detail::Offerings;
    using detail::CourseList;
    using detail::ActivityConfiguration;
    using detail::StudySession;
    using detail::GenerationStatistics;
    using detail::FileSaveStatus;
    using detail::TimetableDataSource;
    using detail::TimetableDataSink;
    using detail::ProgressIndicator;
    using detail::GetEarliestClassTime;
    using detail::GetTotalTimeBlocks;
    using detail::GetTimetableSortCriteria;
    using detail::TimetableManager;
    
}  // end namespace tg

#endif /* LIBTG_INCLUDE_TG_API_HPP_ */
